"""Type fsm.event, version 000"""

from typing import Literal

from gwproto.property_format import (
    HandleName,
    LeftRightDotStr,
    UTCMilliseconds,
    UUID4Str,
)
from pydantic import BaseModel, ConfigDict, model_validator
from typing_extensions import Self


class FsmEvent(BaseModel):
    """
    Finite State Machine Event Command.

    A message sent to a SpaceheatNode wher ethe Node implements a finite state machine. The
    message is intended to be an FSM Events (aka Trigger) that allow a state machine to react
    (by starting a Transition and any side-effect Actions).

    [More info](https://gridworks-protocol.readthedocs.io/en/latest/finite-state-machines.html)
    """

    FromHandle: HandleName
    ToHandle: HandleName
    EventType: LeftRightDotStr
    EventName: str
    TriggerId: UUID4Str
    SendTimeUnixMs: UTCMilliseconds
    TypeName: Literal["fsm.event"] = "fsm.event"
    Version: str = "000"

    model_config = ConfigDict(extra="allow")

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: EventName must belong to the enum selected in the EventType.

        """
        # Implement check for axiom 1"
        return self

    @model_validator(mode="after")
    def check_axiom_2(self) -> Self:
        """
        Axiom 2: FromHandle must be the immediate boss of ToHandle.
        """
        if "multiplexer" in self.ToHandle:
            return self
        immediate_boss = ".".join(self.ToHandle.split(".")[:-1])
        if immediate_boss != self.FromHandle:
            raise ValueError(
                f"FromHandle {self.FromHandle} must be immediate boss of ToHandle {self.ToHandle}"
            )
        return self



"""Type atn.bid, version 001"""

from typing import List, Literal

from gwsproto.enums import MarketPriceUnit, MarketQuantityUnit
from gwproto.property_format import LeftRightDotStr, MarketSlotName
from gwsproto.named_types.price_quantity_unitless import PriceQuantityUnitless
from pydantic import BaseModel, field_validator, model_validator
from typing_extensions import Self


class AtnBid(BaseModel):
    BidderAlias: LeftRightDotStr
    MarketSlotName: MarketSlotName
    PqPairs: List[PriceQuantityUnitless]
    InjectionIsPositive: bool
    PriceUnit: MarketPriceUnit
    QuantityUnit: MarketQuantityUnit
    SignedMarketFeeTxn: str
    TypeName: Literal["atn.bid"] = "atn.bid"
    Version: str = "001"

    @field_validator("SignedMarketFeeTxn")
    @classmethod
    def _check_signed_market_fee_txn(cls, v: str) -> str:
        # supposed to be check_is_algo_msg_pack_encoded, hacked out in sacda
        return v

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: PqPairs PriceMax matches MarketType.
        There is a GridWorks global list of MarketTypes (a GridWorks type), identified by
        their MarketTypeNames (a GridWorks enum).  The MarketType has a PriceMax, which
        must be the first price of the first PriceQuantity pair in PqPairs.
        """
        # Implement check for axiom 1"
        return self



"""Type no.new.contract.warning, version 000"""

from typing import Literal
from pydantic import BaseModel

from gwproto.property_format import LeftRightDotStr, UUID4Str, UTCSeconds
class NoNewContractWarning(BaseModel):
    FromGNodeAlias: LeftRightDotStr
    ContractId: UUID4Str
    GraceEndTimeS: UTCSeconds
    TypeName: Literal["no.new.contract.warning"] = "no.new.contract.warning"
    Version: str = "000"



import uuid
from typing import Literal

from gwproto.property_format import LeftRightDotStr, UUID4Str
from pydantic import BaseModel, Field


class RemainingElec(BaseModel):
    FromGNodeAlias: LeftRightDotStr
    RemainingWattHours: int
    MessageId: UUID4Str = Field(default_factory=lambda: str(uuid.uuid4()))
    TypeName: Literal["remaining.elec"] = "remaining.elec"
    Version: str = "000"



"""Type atn.bid, version 001"""

from typing import List, Literal

from gwsproto.enums import MarketPriceUnit, MarketQuantityUnit, MarketTypeName
from gwproto.property_format import (LeftRightDotStr, MarketSlotName, MarketName, 
                        UUID4Str, UTCMilliseconds)
from gwsproto.named_types.price_quantity_unitless import PriceQuantityUnitless
from pydantic import BaseModel, field_validator, model_validator
from typing_extensions import Self


class AtnBid(BaseModel):
    """Bid from AtomicTNode to MarketMaker
    
    Note: While key information is encoded in MarketSlotName,
    we include individual fields to improve clarity and enable explicit
    validation. The axiom check_axiom_1 ensures all fields are consistent
    with MarketSlotName.
    """
    BidId: UUID4Str
    MarketSlotName: MarketSlotName
    MarketName: MarketName
    MarketType: MarketTypeName
    MarketMakerAlias: LeftRightDotStr
    BidderAlias: LeftRightDotStr
    StartUnixS: int
    DurationMinutes: int
    PqPairs: List[PriceQuantityUnitless]
    InjectionIsPositive: bool
    PriceUnit: MarketPriceUnit
    QuantityUnit: MarketQuantityUnit
    Signature: str  # Algorand signature from bidder's TaTradingRights
    BidTimeMs: UTCMilliseconds
    TypeName: Literal["atn.bid"] = "atn.bid"
    Version: str = "001"

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: MarketSlotName, MarketName, MarketType, MarketMakerAlias,
        StartUnixS, and DurationMinutes are all consistent
        """
        try:
            # Split apart MarketSlotName
            market_slot_parts = self.MarketSlotName.split('.')
            if len(market_slot_parts) < 4:
                raise ValueError("MarketSlotName must have at least 4 parts")
            
            slot_market_service = market_slot_parts[0]  # e|r|d
            slot_market_type = market_slot_parts[1]
            slot_start_s = int(market_slot_parts[-1])
            slot_market_maker = '.'.join(market_slot_parts[2:-1])

            # Split apart MarketName 
            market_parts = self.MarketName.split('.')
            if len(market_parts) < 3:
                raise ValueError("MarketName must have at least 3 parts")
            
            market_service = market_parts[0]  # e|r|d
            market_type_str = market_parts[1]
            market_maker = '.'.join(market_parts[2:])

            # Check consistency
            if slot_market_service != market_service:
                raise ValueError(
                    f"Market service mismatch: {slot_market_service} vs {market_service}"
                )

            if slot_market_type != market_type_str:
                raise ValueError(
                    f"Market type mismatch: {slot_market_type} vs {market_type_str}"
                )

            if slot_market_type != self.MarketType.value:
                raise ValueError(
                    f"MarketType mismatch: {slot_market_type} vs {self.MarketType.value}"
                )

            if slot_market_maker != market_maker:
                raise ValueError(
                    f"Market maker mismatch: {slot_market_maker} vs {market_maker}"
                )

            if market_maker != self.MarketMakerAlias:
                raise ValueError(
                    f"MarketMakerAlias mismatch: {market_maker} vs {self.MarketMakerAlias}"
                )

            if slot_start_s != self.StartUnixS:
                raise ValueError(
                    f"StartUnixS mismatch: {slot_start_s} vs {self.StartUnixS}"
                )

            # Check that the StartUnixS aligns with market duration
            from gw.data_classes.market_type import MarketType
            market_obj = MarketType.by_id.get(self.MarketType)
            if market_obj:
                if self.DurationMinutes != market_obj.duration_minutes:
                    raise ValueError(
                        f"DurationMinutes {self.DurationMinutes} does not match "
                        f"market type {market_obj.name} duration {market_obj.duration_minutes}"
                    )
                if slot_start_s % (market_obj.duration_minutes * 60) != 0:
                    raise ValueError(
                        f"StartUnixS {slot_start_s} not aligned with "
                        f"market duration {market_obj.duration_minutes} minutes"
                    )

        except ValueError as e:
            raise ValueError(f"AtnBid consistency check failed: {str(e)}")
        except Exception as e:
            raise ValueError(f"Error validating AtnBid consistency: {str(e)}")

        return self

    @model_validator(mode="after")
    def check_axiom_2(self) -> Self:
        """Axiom 2: PqPairs are ordered by increasing price
        
        Note: For a bid to buy power (InjectionIsPositive=False), this means
        higher prices lead to lower consumption. For a bid to sell power
        (InjectionIsPositive=True), higher prices lead to higher injection.
        """
        if len(self.PqPairs) < 1:
            raise ValueError("Must have at least one PQ pair")
            
        for i in range(len(self.PqPairs) - 1):
            if self.PqPairs[i].PriceTimes1000 >= self.PqPairs[i+1].PriceTimes1000:
                raise ValueError(
                    f"PQ pairs must have strictly increasing prices. Pair {i} "
                    f"({self.PqPairs[i].PriceTimes1000}) >= Pair {i+1} "
                    f"({self.PqPairs[i+1].PriceTimes1000})"
                )
        return self

    @field_validator("Signature")
    @classmethod
    def _check_signature(cls, v: str) -> str:
        """Validates signature format
        
        Future: Will validate Algorand signature from Terminal Asset Trading Rights 
        NFT. Currently accepts placeholder that follows basic Algorand message pack
        format.
        """
        # TODO: Replace with real Algorand TaTradingRights signature validation
        if not v.startswith("algo_sig_"):
            raise ValueError(
                "Even placeholder signatures must start with 'algo_sig_' "
                "to indicate Algorand signing intent"
            )
        return v


"""Type sieg.target.too.low, version 000"""

from typing import Literal


from gwproto.property_format import  LeftRightDotStr, UTCMilliseconds
from pydantic import BaseModel, StrictInt


class SiegTargetTooLow(BaseModel):
    FromGNodeAlias: LeftRightDotStr
    TargetLwtFx10: StrictInt
    SiegColdFx10: StrictInt
    HeatPumpDeltaTx10: StrictInt
    TimeMs: UTCMilliseconds
    TypeName: Literal["sieg.target.too.low"] = "sieg.target.too.low"
    Version: str = "000"



from typing import List, Literal
from pydantic import BaseModel
from gwproto.property_format import SpaceheatName


class MicroVolts(BaseModel):
    HwUid: str
    AboutNodeNameList: List[SpaceheatName]
    MicroVoltsList: List[int]
    TypeName: Literal["microvolts"] = "microvolts"
    Version: Literal["100"] = "100"



from typing import Optional, Literal
from gwproto.property_format import  UTCMilliseconds, SpaceheatName
from pydantic import BaseModel, field_validator, model_validator
from gwsproto.enums import ContractStatus
from gwsproto.named_types import SlowDispatchContract
from typing_extensions import Self

class SlowContractHeartbeat(BaseModel):
    """Base class for contract lifecycle messages"""
    FromNode: SpaceheatName # either "a" or "s", for Atn or Scada
    Contract: SlowDispatchContract
    PreviousStatus: Optional[ContractStatus] = None
    Status: ContractStatus
    MessageCreatedMs: UTCMilliseconds
    Cause: Optional[str] = None
    IsAuthoritative: bool = True
    WattHoursUsed: Optional[int] = None
    MyDigit: int
    YourLastDigit: Optional[int] = None
    SignedProof: str = "algo_sig_dummy" # For blockchain validation
    TypeName: Literal["slow.contract.heartbeat"] = "slow.contract.heartbeat"
    Version: str = "000"


    def contract_grace_period_minutes(self) -> Literal[5]:
        return 5

    def grace_period_end_s(self) -> int:
        """ 5 minutes after the ContractEndS unless the contract
        was terminated, in which case 5 minutes after termination
        """
        contract_done_s = self.Contract.contract_end_s()
        if self.Status in [ContractStatus.TerminatedByAtn, ContractStatus.TerminatedByScada]:
            contract_done_s = int(self.MessageCreatedMs / 1000)
        return contract_done_s + self.contract_grace_period_minutes() * 60
        # TODO: Add a test for this logic

    @model_validator(mode="after")
    def _check_axiom_1(self) -> Self:
        """Axiom 1: Contracts must be created no later 
        than 10 seconds after StartS"""
        if self.Status in [ContractStatus.Created]:
            time_s = self.MessageCreatedMs / 1000
            if time_s > self.Contract.StartS + 10:
                raise ValueError(
                    f"Axiom 2: Must be {self.Status.value} within 10 seconds of Contract Start. Got {round(time_s - self.Contract.StartS, 2)}"
                )
        return self
    
    @model_validator(mode='after')
    def _check_axiom_2(self) -> Self:
        """Axiom 2 Check authority: Validate authority for status changes"""
        
        # Only Atn can create or confirm
        if self.Status in [ContractStatus.Created, ContractStatus.Confirmed, ContractStatus.TerminatedByAtn]:
            if self.FromNode != "a":
                raise ValueError(f"Only Atn can set status {self.Status}")
            if not self.IsAuthoritative:
                raise ValueError("Atn IsAuthoritative for Created and Confirmed!")
        # Only Scada can mark as received
        if self.Status in [ContractStatus.Received, ContractStatus.TerminatedByScada]:
            if self.FromNode != "s":
                raise ValueError(f"Only Scada can set status {self.Status}")
            if not self.IsAuthoritative:
                raise ValueError("Scada IsAuthoritative for Received!")
        # Active/CompletedSuccess/CompletedFailure are for umpire only
        # For now, treat these as claims by participants
        if self.Status in [ContractStatus.CompletedSuccess,
                          ContractStatus.CompletedFailureByAtn,
                          ContractStatus.CompletedFailureByScada]:
            # Later the umpire will enforce these
            # For now just let participants publish their view
            if self.IsAuthoritative:
                raise ValueError(f"{self.FromNode} is NOT Authoritative for {self.Status}")

        return self

    @model_validator(mode='after')
    def check_axiom_3(self) -> Self:
        """Axiom 3: Cause required for and limited to Terminated and CompletedFailure/Unknown status"""
        needs_cause = self.Status in [ContractStatus.TerminatedByAtn,
                                      ContractStatus.TerminatedByScada, 
                                      ContractStatus.CompletedUnknownOutcome,
                                      ContractStatus.CompletedFailureByAtn,
                                      ContractStatus.CompletedFailureByScada]
        has_cause = self.Cause is not None

        if needs_cause and not has_cause:
            raise ValueError(f"Cause is required for {needs_cause}")
        if not needs_cause and has_cause:
            raise ValueError(f"Cause only valid for {needs_cause} status")
        return self

    @field_validator("FromNode")
    @classmethod
    def _check_from_node(cls, v: str) -> str:
        """
        Axiom 4: FromNode should be 's' (Scada) or 'a' (AtomicTNode)
        """
        if v not in ['s', 'a']:
            raise ValueError(
                f"Axiom 4: FromNode should be 's' (Scada) or 'a' (AtomicTNode). Got {v}"
            )
        return v

    @model_validator(mode='after')
    def check_axiom_5(self) -> Self:
        """
        Future: Will validate Algorand smart contract transaction signature.

        By including YourLastDigit and the previous message's SignedProof,
        this creates an unbreakable chain where:
          - Each party must reference the exact signature from their counterparty's last message
          - The digits create a crosslinked sequence that must match on both sides
          - The signatures create a cryptographic chain that can't be spoofed by one side after-the-fact
        sign({
            "ContractId": "c3cd6c8d-e8d8-4875-8f7b-d3ea13c28693", 
            "FromNode": "a",
            "Status": "Active",
            "MessageCreatedMs": 1740425972511,
            "MyDigit": 3,
            "YourLastDigit": 7,
            "PrevSignedProof": "algo_sig_xyz..."
        })
        Currently accepts placeholder that follows basic Algorand message pack format.
        """
        # TODO: Replace with real Algorand transaction signature validation
        if not self.SignedProof.startswith("algo_sig_"):
            raise ValueError(
                "Even placeholder signatures must start with 'algo_sig_' "
                "to indicate Algorand signing intent"
            )
        return self

    @model_validator(mode='after')
    def check_axiom_6(self) -> Self:
        """Axiom 6: check digits"""
        if self.MyDigit not in range(10):
            raise ValueError(f"MyDigit must be in range(10) not {self.MyDigit}")
        if self.YourLastDigit:
            if self.YourLastDigit not in range(10):
                raise ValueError(f"YourLastDigit must be in range(10) not {self.YourLastDigit}")
        return self

    @model_validator(mode='after')
    def check_axiom_7(self) -> Self:
        """Axiom 7: Scada (and only Scada) sends WattHoursUsed"""
        if self.FromNode == "s":
            if self.WattHoursUsed is None:
                raise ValueError("Scada sends WattHoursUsed")
        else:
            if self.WattHoursUsed:
                raise ValueError("Only Scada sends WattHoursUsed")
        return self







"""Type latest.price, version 000"""

from typing import Literal

from gwsproto.enums import MarketPriceUnit
from gwproto.property_format import LeftRightDotStr, MarketSlotName, UUID4Str
from pydantic import BaseModel, StrictInt


class LatestPrice(BaseModel):
    FromGNodeAlias: LeftRightDotStr
    PriceTimes1000: StrictInt
    PriceUnit: MarketPriceUnit
    MarketSlotName: MarketSlotName
    MessageId: UUID4Str
    TypeName: Literal["latest.price"] = "latest.price"
    Version: str = "000"



# Literal Enum:
#  - no additional values can be added over time.
#  - Sent as-is, not in hex symbol
from enum import auto
from typing import List

from gw.enums import GwStrEnum


class ChangeKeepSend(GwStrEnum):
    """
    HpLoop relay - change to send more or send less
    """

    ChangeToKeepLess = auto()
    ChangeToKeepMore = auto()

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]

    @classmethod
    def default(cls) -> "ChangeKeepSend":
        return cls.ChangeToKeepLess

    @classmethod
    def enum_name(cls) -> str:
        return "change.keep.send"



"""Type scada.params, version 002"""

from typing import Literal, Optional

from gwproto.property_format import (
    LeftRightDotStr,
    SpaceheatName,
    UTCMilliseconds,
    UUID4Str,
)
from gwsproto.named_types.ha1_params import Ha1Params
from pydantic import BaseModel, ConfigDict


class ScadaParams(BaseModel):
    FromGNodeAlias: LeftRightDotStr
    FromName: SpaceheatName
    ToName: SpaceheatName
    UnixTimeMs: UTCMilliseconds
    MessageId: UUID4Str
    NewParams: Optional[Ha1Params] = None
    OldParams: Optional[Ha1Params] = None
    TypeName: Literal["scada.params"] = "scada.params"
    Version: str = "002"

    model_config = ConfigDict(extra="allow")



"""Type reset.hp.keep.value, version 000"""
import time
from typing import Literal

from gwproto.property_format import HandleName, UTCMilliseconds
from pydantic import BaseModel, Field, StrictInt


class ResetHpKeepValue(BaseModel):
    """
    Used to change the HpKeepSeconds - an integrated value meant to represent the 
    position of the Siegenthaler Valve from 0 ("fully send") to 100 ("fully keep") - 
    WITHOUT changing the valve
    """
    FromHandle: HandleName
    ToHandle: HandleName
    HpKeepSecondsTimes10: StrictInt
    CreatedMs: UTCMilliseconds = Field(default_factory=lambda: int(time.time() * 1000))
    TypeName: Literal["reset.hp.keep.value"] = "reset.hp.keep.value"
    Version: str = "001"



# Literal Enum:
#  - no additional values can be added over time.
#  - Sent as-is, not in hex symbol
from enum import auto
from typing import List

from gw.enums import GwStrEnum


class HpLoopKeepSend(GwStrEnum):
    """
    """

    SendLess = auto()
    SendMore = auto()

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]

    @classmethod
    def default(cls) -> "HpLoopKeepSend":
        return cls.SendMore

    @classmethod
    def enum_name(cls) -> str:
        return "hp.loop.keep.send"



from enum import auto
from typing import List

from gw.enums import GwStrEnum


class HomeAloneTopState(GwStrEnum):
    """
    
    Values:
      - Dormant
      - UsingBackupOnpeak
      - Normal
      - ScadaBlind
      - Monitor

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#homealonetopstate)
    """

    Dormant = auto()
    UsingBackupOnpeak = auto()
    Normal = auto()
    ScadaBlind = auto()
    Monitor = auto()

    @classmethod
    def default(cls) -> "HomeAloneTopState":
        return cls.Dormant

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "home.alone.top.state"

    @classmethod
    def enum_version(cls) -> str:
        return "001"



"""Type single.machine.state, version 000"""

from typing import Literal, Optional

from gwproto.property_format import HandleName, LeftRightDotStr, UTCMilliseconds
from pydantic import BaseModel, model_validator
from typing_extensions import Self


class SingleMachineState(BaseModel):
    """
    Intra-scada comms type for sharing the current state of a state machine.
    """

    MachineHandle: HandleName
    StateEnum: LeftRightDotStr
    State: str
    UnixMs: UTCMilliseconds
    Cause: Optional[str] = None
    TypeName: Literal["single.machine.state"] = "single.machine.state"
    Version: str = "000"

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: If StateEnum is a recongized enum, then State is a value of that enum.

        """
        # Implement check for axiom 1"
        return self



"""Type wake.up, version 000"""

from typing import Literal

from gwproto.property_format import SpaceheatName
from pydantic import BaseModel


class WakeUp(BaseModel):
    ToName: SpaceheatName
    TypeName: Literal["wake.up"] = "wake.up"
    Version: str = "000"



"""Type go.dormant, version 000"""

from typing import Literal
from gwproto.property_format import SpaceheatName
from pydantic import BaseModel


class GoDormant(BaseModel):
    ToName: SpaceheatName
    TypeName: Literal["go.dormant"] = "go.dormant"
    Version: str = "001"



from typing import ClassVar, Dict, List, Literal
from pydantic import BaseModel
from gwproto.enums import TelemetryName
from gwproto.property_format import SpaceheatName
DEFAULT_ANALOG_READER = "analog-temp"


class ChannelStub(BaseModel):
    Name: SpaceheatName
    AboutNodeName: SpaceheatName
    TelemetryName: TelemetryName
    InPowerMetering: bool = False

    def __hash__(self) -> int:
        return hash(self.Name)

class ZoneNodes:
    def __init__(self, zone: str, idx: int) -> None:
        zone_name = f"zone{idx + 1}-{zone}".lower()
        self.zone_name = zone_name
        self.stat = f"{zone_name}-stat"
        self.whitewire=f"zone{idx + 1}-{zone}-whitewire"

    def __repr__(self) -> str:
        return f"Zone {self.zone_name} and stat {self.stat}"


class TankNodes:
    def __init__(self, tank_name: str) -> None:
        self.reader = tank_name
        self.depth1 = f"{tank_name}-depth1"
        self.depth2 = f"{tank_name}-depth2"
        self.depth3 = f"{tank_name}-depth3"
        self.depth4 = f"{tank_name}-depth4"

    def __repr__(self) -> str:
        return f"{self.reader} reads {self.depth1}, {self.depth2}, {self.depth3}, {self.depth4}"


class ZoneChannelNames:
    def __init__(self, zone: str, idx: int) -> None:
        self.zone_name = f"zone{idx}-{zone}".lower()
        self.stat_name = f"{self.zone_name}-stat"
        self.temp = f"{self.zone_name}-temp"
        self.set = f"{self.zone_name}-set"
        self.state = f"{self.zone_name}-state"
        self.whitewire_pwr=f"zone{idx}-{zone}-whitewire-pwr"

    def __repr__(self) -> str:
        return f"Channels: {self.temp}, {self.set}, {self.state}, {self.whitewire_pwr}"


class TankChannelNames:
    def __init__(self, tank_name: str) -> None:
        self.depth1 = f"{tank_name}-depth1"
        self.depth2 = f"{tank_name}-depth2"
        self.depth3 = f"{tank_name}-depth3"
        self.depth4 = f"{tank_name}-depth4"


class House0RelayIdx:
    vdc: Literal[1] = 1
    tstat_common: Literal[2] = 2
    store_charge_disharge: Literal[3] = 3
    hp_failsafe: Literal[5] = 5
    hp_scada_ops: Literal[6] = 6
    thermistor_common: Literal[7] = 7
    aquastat_ctrl: Literal[8] = 8
    store_pump_failsafe: Literal[9] = 9
    boiler_scada_ops: Literal[10] = 10
    primary_pump_ops: Literal[11] = 11
    primary_pump_failsafe: Literal[12] = 12
    hp_loop_on_off: Literal[14] = 14
    hp_loop_keep_send: Literal[15] = 15
    # pattern: zone1 failsafe is 17, zone2 ops is 18 etc
    base_stat: Literal[17] = 17

class H0N:
    # core actors
    atn = "a"
    atomic_ally = "aa"
    primary_scada = "s"
    secondary_scada = "s2"
    home_alone = "h"
    home_alone_normal = "n"
    home_alone_onpeak_backup = "onpeak-backup"
    home_alone_scada_blind = "scada-blind"
    primary_power_meter = "power-meter"
    admin = "admin"  # used when starter scripts take control
    auto = "auto"  # Finite State Machine responsible for homealone <-> atn transition
    analog_temp = "analog-temp"
    relay_multiplexer = "relay-multiplexer"
    synth_generator = "synth-generator"
    zero_ten_out_multiplexer = "dfr-multiplexer"
    hp_boss = "hp-boss"
    sieg_loop = "sieg-loop"

    # core power-metered nodes
    hp_odu = "hp-odu"
    hp_idu = "hp-idu"
    dist_pump = "dist-pump"
    primary_pump = "primary-pump"
    store_pump = "store-pump"

    # core temperatures
    dist_swt = "dist-swt"
    dist_rwt = "dist-rwt"
    hp_lwt = "hp-lwt"
    hp_ewt = "hp-ewt"
    buffer_hot_pipe = "buffer-hot-pipe"
    buffer_cold_pipe = "buffer-cold-pipe"
    store_hot_pipe = "store-hot-pipe"
    store_cold_pipe = "store-cold-pipe"
    oat = "oat"
    sieg_cold = "sieg-cold"
    buffer = TankNodes("buffer")
    tank: ClassVar[Dict[int, TankNodes]] = {}
    zone: ClassVar[Dict[str, ZoneNodes]] = {}

    # core flow-metered nodes
    dist_flow = "dist-flow"
    primary_flow = "primary-flow"
    store_flow = "store-flow"
    sieg_flow = "sieg-flow"

    # relay nodes
    vdc_relay: Literal["relay1"] = "relay1"
    tstat_common_relay: Literal["relay2"] = "relay2"
    store_charge_discharge_relay: Literal["relay3"] = "relay3"
    hp_failsafe_relay: Literal["relay5"] = "relay5"
    hp_scada_ops_relay: Literal["relay6"] = "relay6"
    thermistor_common_relay: Literal["relay7"] = "relay7"
    aquastat_ctrl_relay: Literal["relay8"] = "relay8"
    store_pump_failsafe: Literal["relay9"] = "relay9"

    boiler_scada_ops: Literal["relay10"] = "relay10"
    primary_pump_scada_ops: Literal["relay11"] = "relay11"
    primary_pump_failsafe: Literal["relay12"] = "relay12"
    hp_loop_on_off: Literal["relay14"] = "relay14"
    hp_loop_keep_send: Literal["relay15"] = "relay15"

    # zero ten output
    dist_010v = "dist-010v"
    primary_010v = "primary-010v"
    store_010v = "store-010v"
    hubitat = "hubitat"

    # finite state machines
    pico_cycler = "pico-cycler"

    def __init__(self, total_store_tanks: int, zone_list: List[str]) -> None:
        for i in range(total_store_tanks):
            self.tank[i + 1] = TankNodes(f"tank{i + 1}")
        for i in range(len(zone_list)):
            self.zone[zone_list[i]] = ZoneNodes(zone=zone_list[i], idx=i)


class House0StartHandles:
    scada = "h.s"
    admin = "admin"
    home_alone = "h"
    relay_multiplexer = "admin.relay-multiplexer"


class H0CN:
    # Power Channels
    hp_odu_pwr = f"{H0N.hp_odu}-pwr"
    hp_idu_pwr = f"{H0N.hp_idu}-pwr"
    dist_pump_pwr = f"{H0N.dist_pump}-pwr"
    primary_pump_pwr = f"{H0N.primary_pump}-pwr"
    store_pump_pwr = f"{H0N.store_pump}-pwr"

    # Temperature Channels
    dist_swt = H0N.dist_swt
    dist_rwt = H0N.dist_rwt
    hp_lwt = H0N.hp_lwt
    hp_ewt = H0N.hp_ewt
    store_hot_pipe = H0N.store_hot_pipe
    store_cold_pipe = H0N.store_cold_pipe
    buffer_hot_pipe = H0N.buffer_hot_pipe
    buffer_cold_pipe = H0N.buffer_cold_pipe
    oat = H0N.oat
    sieg_cold = H0N.sieg_cold
    buffer = TankChannelNames("buffer")
    tank: ClassVar[Dict[int, TankChannelNames]] = {}
    zone: ClassVar[Dict[int, ZoneChannelNames]] = {}

    # Flow Channels
    dist_flow = H0N.dist_flow
    primary_flow = H0N.primary_flow
    store_flow = H0N.store_flow
    sieg_flow = H0N.sieg_flow
    dist_flow_hz = f"{H0N.dist_flow}-hz"
    primary_flow_hz = f"{H0N.primary_flow}-hz"
    store_flow_hz = f"{H0N.store_flow}-hz"

    # Synth Channels
    required_energy = "required-energy"
    usable_energy = "usable-energy"
    hp_keep_seconds_x_10 = "hp-keep-seconds-x-10"

    # relay state channels
    vdc_relay_state: Literal["vdc-relay1"] = "vdc-relay1"
    tstat_common_relay_state: Literal["tstat-common-relay2"] = "tstat-common-relay2"
    charge_discharge_relay_state: Literal["charge-discharge-relay3"] = "charge-discharge-relay3"
    hp_failsafe_relay_state = f"hp-failsafe-{H0N.hp_failsafe_relay}"
    thermistor_common_relay_state = f"thermistor-common-{H0N.thermistor_common_relay}"
    hp_scada_ops_relay_state = f"hp-scada-ops-{H0N.hp_scada_ops_relay}"
    aquastat_ctrl_relay_state = f"aquastat-ctrl-{H0N.aquastat_ctrl_relay}"
    store_pump_failsafe_relay_state = f"store-pump-failsafe-{H0N.store_pump_failsafe}"
    boiler_scada_ops_relay_state = f"boiler-scada_ops-{H0N.boiler_scada_ops}"
    primary_pump_scada_ops_relay_state = (
        f"primary-pump-scada-ops-{H0N.primary_pump_scada_ops}"
    )
    primary_pump_failsafe_relay_state = (
        f"primary-pump-failsafe-{H0N.primary_pump_failsafe}"
    )

    hp_loop_on_off_relay_state = f"hp-loop-on-off-{H0N.hp_loop_on_off}"
    hp_loop_keep_send_relay_state = f"hp-loop-keep-send-{H0N.hp_loop_keep_send}"

    # 010V output state (as declared by entity sending, not reading)
    dist_010v = "dist-010v"
    primary_010v = "primary-010v"
    store_010v = "store-010v"

    def __init__(self, total_store_tanks: int, zone_list: List[str]) -> None:
        for i in range(total_store_tanks):
            self.tank[i + 1] = TankChannelNames(f"tank{i + 1}")
        for i in range(len(zone_list)):
            self.zone[i + 1] = ZoneChannelNames(zone=zone_list[i], idx=i + 1)

    def channel_stubs(self) -> Dict[str, ChannelStub]:
        d = {
            self.hp_odu_pwr: ChannelStub(
                Name=self.hp_odu_pwr,
                AboutNodeName=H0N.hp_odu,
                TelemetryName=TelemetryName.PowerW,
                InPowerMetering=True,
            ),
            self.hp_idu_pwr: ChannelStub(
                Name=self.hp_idu_pwr,
                AboutNodeName=H0N.hp_idu,
                TelemetryName=TelemetryName.PowerW,
                InPowerMetering=True,
            ),
            self.dist_pump_pwr: ChannelStub(
                Name=self.dist_pump_pwr,
                AboutNodeName=H0N.dist_pump,
                TelemetryName=TelemetryName.PowerW,
            ),
            self.primary_pump_pwr: ChannelStub(
                Name=self.primary_pump_pwr,
                AboutNodeName=H0N.primary_pump,
                TelemetryName=TelemetryName.PowerW,
            ),
            self.store_pump_pwr: ChannelStub(
                Name=self.store_pump_pwr,
                AboutNodeName=H0N.store_pump,
                TelemetryName=TelemetryName.PowerW,
            ),
            self.dist_swt: ChannelStub(
                Name=self.dist_swt,
                AboutNodeName=H0N.dist_swt,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            ),
            self.dist_rwt: ChannelStub(
                Name=self.dist_rwt,
                AboutNodeName=H0N.dist_rwt,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            ),
            self.hp_lwt: ChannelStub(
                Name=self.hp_lwt,
                AboutNodeName=H0N.hp_lwt,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            ),
            self.hp_ewt: ChannelStub(
                Name=self.hp_ewt,
                AboutNodeName=H0N.hp_ewt,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            ),
            self.store_hot_pipe: ChannelStub(
                Name=self.store_hot_pipe,
                AboutNodeName=H0N.store_hot_pipe,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            ),
            self.store_cold_pipe: ChannelStub(
                Name=self.store_cold_pipe,
                AboutNodeName=H0N.store_cold_pipe,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            ),
            self.buffer_hot_pipe: ChannelStub(
                Name=self.buffer_hot_pipe,
                AboutNodeName=H0N.buffer_hot_pipe,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            ),
            self.buffer_cold_pipe: ChannelStub(
                Name=self.buffer_cold_pipe,
                AboutNodeName=H0N.buffer_cold_pipe,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            ),
            self.oat: ChannelStub(
                Name=self.oat,
                AboutNodeName=H0N.oat,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            ),
            self.dist_flow: ChannelStub(
                Name=self.dist_flow,
                AboutNodeName=H0N.dist_flow,
                TelemetryName=TelemetryName.GpmTimes100,
            ),
            self.primary_flow: ChannelStub(
                Name=self.primary_flow,
                AboutNodeName=H0N.primary_flow,
                TelemetryName=TelemetryName.GpmTimes100,
            ),
            self.store_flow: ChannelStub(
                Name=self.store_flow,
                AboutNodeName=H0N.store_flow,
                TelemetryName=TelemetryName.GpmTimes100,
            ),
            self.buffer.depth1: ChannelStub(
                Name=self.buffer.depth1,
                AboutNodeName=H0N.buffer.depth1,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            ),
            self.buffer.depth2: ChannelStub(
                Name=self.buffer.depth2,
                AboutNodeName=H0N.buffer.depth2,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            ),
            self.buffer.depth3: ChannelStub(
                Name=self.buffer.depth3,
                AboutNodeName=H0N.buffer.depth3,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            ),
            self.buffer.depth4: ChannelStub(
                Name=self.buffer.depth4,
                AboutNodeName=H0N.buffer.depth4,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            ),
        }
        for i in self.tank:
            d[self.tank[i].depth1] = ChannelStub(
                Name=self.tank[i].depth1,
                AboutNodeName=self.tank[i].depth1,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            )
            d[self.tank[i].depth2] = ChannelStub(
                Name=self.tank[i].depth2,
                AboutNodeName=self.tank[i].depth2,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            )
            d[self.tank[i].depth3] = ChannelStub(
                Name=self.tank[i].depth3,
                AboutNodeName=self.tank[i].depth3,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            )
            d[self.tank[i].depth4] = ChannelStub(
                Name=self.tank[i].depth4,
                AboutNodeName=self.tank[i].depth4,
                TelemetryName=TelemetryName.WaterTempCTimes1000,
            )
        for i in self.zone:
            d[self.zone[i].temp] = ChannelStub(
                    Name=self.zone[i].temp,
                    AboutNodeName=self.zone[i].zone_name,
                    TelemetryName=TelemetryName.AirTempFTimes1000,
                ),
    
            d[self.zone[i].set] = ChannelStub(
                    Name=self.zone[i].temp,
                    AboutNodeName=self.zone[i].stat_name,
                    TelemetryName=TelemetryName.AirTempFTimes1000,
                ),
            d[self.zone[i].set] = ChannelStub(
                    Name=self.zone[i].state,
                    AboutNodeName=self.zone[i].stat_name,
                    TelemetryName=TelemetryName.ThermostatState,
                ),
            d[self.zone[i].whitewire] = ChannelStub(
                    Name=self.zone[i].state,
                    AboutNodeName=self.zone[i].stat_name,
                    TelemetryName=TelemetryName.ThermostatState,
                ),
            
        return d



"""Type glitch, version 000"""
import time
from typing import Literal

from gwsproto.enums import LogLevel
from gwproto.property_format import LeftRightDotStr, SpaceheatName, UTCMilliseconds
from pydantic import BaseModel, Field


class Glitch(BaseModel):
    FromGNodeAlias: LeftRightDotStr
    Node: SpaceheatName
    Type: LogLevel
    Summary: str
    Details: str
    CreatedMs: UTCMilliseconds = Field(default_factory=lambda: int(time.time() * 1000))
    TypeName: Literal["glitch"] = "glitch"
    Version: str = "000"




"""Type set.lwt.control.params, version 000"""
import time
from typing import Literal

from gwproto.property_format import HandleName, UTCMilliseconds
from pydantic import BaseModel, PositiveInt, Field
from typing_extensions import Self


class SetLwtControlParams(BaseModel):
    FromHandle: HandleName
    ToHandle: HandleName
    ProportionalGain: float
    IntegralGain: float
    DerivativeGain: float
    ControlIntervalSeconds: PositiveInt
    T1: PositiveInt
    T2: PositiveInt
    CreatedMs:  UTCMilliseconds = Field(default_factory=lambda: int(time.time() * 1000))
    TypeName: Literal["set.lwt.control.params"] = "set.lwt.control.params"
    Version: str = "000"


"""Type ha1.params, version 001"""

from typing import Literal

from pydantic import BaseModel, StrictInt


class Ha1Params(BaseModel):
    AlphaTimes10: StrictInt
    BetaTimes100: StrictInt
    GammaEx6: StrictInt
    IntermediatePowerKw: float
    IntermediateRswtF: StrictInt
    DdPowerKw: float
    DdRswtF: StrictInt
    DdDeltaTF: StrictInt
    HpMaxKwTh: float
    MaxEwtF: StrictInt
    LoadOverestimationPercent: StrictInt
    StratBossDist010: StrictInt
    TypeName: Literal["ha1.params"] = "ha1.params"
    Version: str = "003"



"""Type price.quantity.unitless, version 000"""

from typing import Literal

from pydantic import BaseModel, StrictInt


class PriceQuantityUnitless(BaseModel):
    PriceTimes1000: StrictInt
    QuantityTimes1000: StrictInt
    TypeName: Literal["price.quantity.unitless"] = "price.quantity.unitless"
    Version: str = "000"



"""Type dispatch.contract.go.live, version 000"""

from typing import Literal

from gwproto.property_format import LeftRightDotStr
from pydantic import BaseModel, field_validator


class DispatchContractGoLive(BaseModel):
    """
    Triggers DispatchContract GoLive.

    Sent by the Atn to its SCADA when they share an existing DispatchContract. If the SCADA
    is in HomeAlone and gets this message, it will move into Atn mode.
    """

    FromGNodeAlias: LeftRightDotStr
    BlockchainSig: str
    TypeName: Literal["dispatch.contract.go.live"] = "dispatch.contract.go.live"
    Version: str = "000"

    @field_validator("BlockchainSig")
    @classmethod
    def _check_blockchain_sig(cls, v: str) -> str:
        # Add later check_is_algo_msg_pack_encoded(v)
        return v



from typing import Any, Literal

from gwproto.messages.event import EventBase
from gwsproto.named_types.remaining_elec import RemainingElec


class RemainingElecEvent(EventBase):
    Remaining: RemainingElec
    TypeName: Literal["remaining.elec.event"] = "remaining.elec.event"
    Version: str = "000"

    def __init__(self, **data: dict[str, Any]) -> None:
        super().__init__(**data)



"""Type dispatch.contract.go.dormant, version 000"""

from typing import Literal

from gwproto.property_format import LeftRightDotStr
from pydantic import BaseModel, field_validator


class DispatchContractGoDormant(BaseModel):
    """ """

    FromGNodeAlias: LeftRightDotStr
    BlockchainSig: str
    TypeName: Literal["dispatch.contract.go.dormant"] = "dispatch.contract.go.dormant"
    Version: str = "000"

    @field_validator("BlockchainSig")
    @classmethod
    def _check_blockchain_sig(cls, v: str) -> str:
        # Add later: check_is_algo_msg_pack_encoded(v)
        return v



"""
GridWorks Enums used in scada, the Application Shared Language (ASL) used by SCADA
devices and AtomicTNodes to communicate with each other. These enums play a specific structural
role as semantic "glue" within ASLs.

Application Shared Languages are an evolution of the concept of Application Programming Interfaces.
In a nutshell, an API can be viewed as a rather restricted version of an SAL, where only one application
has anything complex/interesting to say and, in general, the developers/owners of that application
have sole responsibility for managing the versioning and changing of that API. Note also that SALs
do not make any a priori assumption about the relationship (i.e. the default client/server for an API)
or the message delivery mechanism (i.e. via default GET/POST to RESTful URLs). For more information
on these ideas:
  - [GridWorks Enums](https://gridwork-type-registry.readthedocs.io/en/latest/types.html)
  - [GridWorks Types](https://gridwork-type-registry.readthedocs.io/en/latest/types.html)
  - [ASLs](https://gridwork-type-registry.readthedocs.io/en/latest/asls.html)
 """

from gw.enums import MarketTypeName
from gwsproto.enums.atomic_ally_state import AtomicAllyState
from gwsproto.enums.contract_status import ContractStatus
from gwsproto.enums.change_keep_send import ChangeKeepSend
from gwsproto.enums.flow_manifold_variant import FlowManifoldVariant
from gwsproto.enums.home_alone_strategy import HomeAloneStrategy
from gwsproto.enums.home_alone_top_state import HomeAloneTopState
from gwsproto.enums.hp_model import HpModel
from gwsproto.enums.hp_loop_keep_send import HpLoopKeepSend
from gwsproto.enums.log_level import LogLevel
from gwsproto.enums.main_auto_event import MainAutoEvent
from gwsproto.enums.main_auto_state import MainAutoState
from gwsproto.enums.market_price_unit import MarketPriceUnit
from gwsproto.enums.market_quantity_unit import MarketQuantityUnit
from gwsproto.enums.pico_cycler_event import PicoCyclerEvent
from gwsproto.enums.pico_cycler_state import PicoCyclerState
from gwsproto.enums.top_event import TopEvent
from gwsproto.enums.top_state import TopState
from gwsproto.enums.turn_hp_on_off import TurnHpOnOff


__all__ = [
    "AtomicAllyState",
    "ContractStatus",
    "MarketTypeName",
    "HpModel",
    "ChangeKeepSend",
    "FlowManifoldVariant",
    "HomeAloneStrategy",
    "HomeAloneTopState",  # [home.alone.top.state.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#homealonetopstate)
    "HpLoopKeepSend",
    "LogLevel",  # [log.level.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#loglevel)
    "MainAutoEvent",  # [main.auto.event.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#mainautoevent)
    "MainAutoState",  # [main.auto.state.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#mainautostate)
    "MarketPriceUnit",  # [market.price.unit.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#marketpriceunit)
    "MarketQuantityUnit",  # [market.quantity.unit.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#marketquantityunit)
    "PicoCyclerEvent",  # [pico.cycler.event.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#picocyclerevent)
    "PicoCyclerState",  # [pico.cycler.state.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#picocyclerstate)
    "TopEvent",  # [top.event.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#topevent)
    "TopState",  # [top.state.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#topstate)
    "TurnHpOnOff",
]



from enum import auto
from typing import List

from gw.enums import GwStrEnum


class HomeAloneStrategy(GwStrEnum):
    """
    Values:
      - WinterTou
      - ShoulderTou
      - Summer

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#homealonetopstate)
    """
    WinterTou = auto()
    ShoulderTou = auto()
    Summer = auto()
  
    @classmethod
    def default(cls) -> "HomeAloneStrategy":
        return cls.WinterTou

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "home.alone.strategy"

    @classmethod
    def enum_version(cls) -> str:
        return "001"



"""Type snapshot.spaceheat, version 002"""

from typing import List, Literal

from pydantic import BaseModel

from gwproto.named_types.single_reading import SingleReading
from gwproto.property_format import (
    LeftRightDotStr,
    UTCMilliseconds,
    UUID4Str,
)
from gwsproto.named_types import SingleMachineState


class SnapshotSpaceheat(BaseModel):
    """
    Snapshot.

    Collection of all the latest measurements (timestamped) captured by the SCADA for all of
    its data channels. Add LatestStateList
    """

    FromGNodeAlias: LeftRightDotStr
    FromGNodeInstanceId: UUID4Str
    SnapshotTimeUnixMs: UTCMilliseconds
    LatestReadingList: List[SingleReading]
    LatestStateList: List[SingleMachineState]
    TypeName: Literal["snapshot.spaceheat"] = "snapshot.spaceheat"
    Version: str = "003"



"""Type heating.forecast, version 000"""
import time
from typing import List, Literal

from gwproto.property_format import LeftRightDotStr, UTCSeconds, UUID4Str
from pydantic import BaseModel, Field, model_validator
from typing_extensions import Self


class HeatingForecast(BaseModel):
    FromGNodeAlias: LeftRightDotStr
    Time: List[UTCSeconds]
    AvgPowerKw: List[float]
    RswtF: List[float]
    RswtDeltaTF: List[float]
    WeatherUid: UUID4Str
    ForecastCreatedS: UTCSeconds = Field(default_factory=lambda: int(time.time()))
    TypeName: Literal["heating.forecast"] = "heating.forecast"
    Version: str = "000"

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: Length of all hte lists must be the same.

        """
        if len(self.AvgPowerKw) != len(self.Time):
            raise ValueError(
                f"self.AvgPowerKw should be a list of length {len(self.Time)}!"
            )
        if len(self.RswtF) != len(self.Time):
            raise ValueError(f"self.RswtF should be a list of length {len(self.Time)}!")
        if len(self.RswtDeltaTF) != len(self.Time):
            raise ValueError(
                f"self.RswtDeltaTFshould be a list of length {len(self.Time)}!"
            )
        return self

    @model_validator(mode="after")
    def check_axiom_2(self) -> Self:
        """
        Axiom 2: ForecastCreatedS is less than the first second in Time
        """
        if len(self.Time) > 0:
            if self.ForecastCreatedS > self.Time[0]:
                raise ValueError(
                    "Forecast should not be created after beginning of forecasted time!"
                )

        return self



"""Type admin.dispatch, version 000"""

from typing import Literal

from gwsproto.named_types.fsm_event import FsmEvent
from pydantic import BaseModel, StrictInt


class AdminDispatch(BaseModel):
    DispatchTrigger: FsmEvent
    TimeoutSeconds: StrictInt
    TypeName: Literal["admin.dispatch"] = "admin.dispatch"
    Version: str = "000"



"""Type hack.oil.on, version 000"""

from typing import Literal

from pydantic import BaseModel


class HackOilOn(BaseModel):
    TypeName: Literal["hack.oil.on"] = "hack.oil.on"
    Version: str = "000"



"""Type pico.missing, version 000"""

from typing import Literal

from gwproto.property_format import SpaceheatName
from pydantic import BaseModel


class PicoMissing(BaseModel):
    ActorName: SpaceheatName
    PicoHwUid: str
    TypeName: Literal["pico.missing"] = "pico.missing"
    Version: str = "000"



from typing import Literal
from pydantic import BaseModel, field_validator, model_validator
from typing_extensions import Self
from gw.enums import MarketTypeName
from gwproto.property_format import UUID4Str, MarketName,MarketSlotName, LeftRightDotStr, UTCMilliseconds

class MarketMakerAck(BaseModel):
    """Acknowledgement of bid
    
    Note: While all necessary information is encoded in MarketSlotName,
    we include individual fields to improve clarity and enable explicit
    validation. The axiom check_axiom_1 ensures all fields are consistent
    with MarketSlotName.
    """
    MarketSlotName: MarketSlotName
    MarketName: MarketName
    MarketType: MarketTypeName
    MarketMakerAlias: LeftRightDotStr
    BidderAlias: LeftRightDotStr
    StartUnixS: int
    DurationMinutes: int
    ContractId: UUID4Str  # BidId from original AtnBid
    Signature: str
    AckTimeMs: UTCMilliseconds
    TypeName: Literal["market.maker.ack"] = "market.maker.ack"
    Version: str = "000"

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1:  MarketSlotName, MarketName, MarketType, MarketMakerAlias,
        StartUnixS, and DurationMinutes are all consistent.

        MarketSlotName format is: e|r|d.{market_type}.{market_maker_alias}.{start_s}
        MarketName format is: e|r|d.{market_type}.{market_maker_alias}
        """
        try:
            # Split apart MarketSlotName
            market_slot_parts = self.MarketSlotName.split('.')
            if len(market_slot_parts) < 4:
                raise ValueError("MarketSlotName must have at least 4 parts")
            
            slot_market_service = market_slot_parts[0]  # e|r|d
            slot_market_type = market_slot_parts[1]
            slot_start_s = int(market_slot_parts[-1])
            slot_market_maker = '.'.join(market_slot_parts[2:-1])

            # Split apart MarketName 
            market_parts = self.MarketName.split('.')
            if len(market_parts) < 3:
                raise ValueError("MarketName must have at least 3 parts")
            
            market_service = market_parts[0]  # e|r|d
            market_type_str = market_parts[1]
            market_maker = '.'.join(market_parts[2:])

            # Check consistency
            if slot_market_service != market_service:
                raise ValueError(
                    f"Market service mismatch: {slot_market_service} vs {market_service}"
                )

            if slot_market_type != market_type_str:
                raise ValueError(
                    f"Market type mismatch: {slot_market_type} vs {market_type_str}"
                )

            if slot_market_type != self.MarketType.value:
                raise ValueError(
                    f"MarketType mismatch: {slot_market_type} vs {self.MarketType.value}"
                )

            if slot_market_maker != market_maker:
                raise ValueError(
                    f"Market maker mismatch: {slot_market_maker} vs {market_maker}"
                )

            if market_maker != self.MarketMakerAlias:
                raise ValueError(
                    f"MarketMakerAlias mismatch: {market_maker} vs {self.MarketMakerAlias}"
                )

            if slot_start_s != self.StartUnixS:
                raise ValueError(
                    f"StartUnixS mismatch: {slot_start_s} vs {self.StartUnixS}"
                )

            # Check that the StartUnixS aligns with market duration
            from gw.data_classes.market_type import MarketType
            market_obj = MarketType.by_id.get(self.MarketType)
            if market_obj:
                if self.DurationMinutes != market_obj.duration_minutes:
                    raise ValueError(
                        f"DurationMinutes {self.DurationMinutes} does not match "
                        f"market type {market_obj.name} duration {market_obj.duration_minutes}"
                    )
                if slot_start_s % (market_obj.duration_minutes * 60) != 0:
                    raise ValueError(
                        f"StartUnixS {slot_start_s} not aligned with "
                        f"market duration {market_obj.duration_minutes} minutes"
                    )

        except ValueError as e:
            raise ValueError(f"MarketMakerAck consistency check failed: {str(e)}")
        except Exception as e:
            raise ValueError(f"Error validating MarketMakerAck consistency: {str(e)}")

        return self

    @field_validator("Signature")
    @classmethod
    def _check_signature(cls, v: str) -> str:
        """
        Future: Will validate Algorand smart contract transaction signature.
        Currently accepts placeholder that follows basic Algorand message pack format.
        """
        # TODO: Replace with real Algorand transaction signature validation
        if not v.startswith("algo_sig_"):
            raise ValueError(
                "Even placeholder signatures must start with 'algo_sig_' "
                "to indicate Algorand signing intent"
            )
        return v



"""Type channel.flatlined, version 000"""

from typing import Literal

from gwproto.named_types import DataChannelGt
from gwproto.property_format import SpaceheatName
from pydantic import BaseModel, model_validator
from typing_extensions import Self


class ChannelFlatlined(BaseModel):
    FromName: SpaceheatName
    Channel: DataChannelGt
    TypeName: Literal["channel.flatlined"] = "channel.flatlined"
    Version: str = "000"

    @model_validator(mode="after")
    def check_axiom_2(self) -> Self:
        """
        Axiom 1:FromName must be Channel.CapturedByNodeName
        """
        if self.Channel.CapturedByNodeName != self.FromName:
            raise ValueError(
                "Axiom 1: FromName must be Channel.CapturedByNodeName"
                f" FromName {self.FromName} != Channel.CapturedByName {self.Channel.CapturedByNodeName}"
            )
        return self



from enum import auto
from typing import List, Optional
from gw.enums import GwStrEnum


class LogLevel(GwStrEnum):
    """
    
    Values:
      - Critical
      - Error
      - Warning
      - Info
      - Debug
      - Trace

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#loglevel)
    """

    Critical = auto()
    Error = auto()
    Warning = auto()
    Info = auto()
    Debug = auto()
    Trace = auto()

    @classmethod
    def default(cls) -> "LogLevel":
        return cls.Info

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "log.level"

    @classmethod
    def enum_version(cls) -> str:
        return "000"

    @classmethod
    def highest_level(cls, levels: List["LogLevel"]) -> Optional["LogLevel"]:
        """
        Returns the highest level from a list of LogLevels.
        If the list is empty, returns None.
        """
        if not levels:
            return None
        return min(levels, key=lambda level: list(cls).index(level))



"""Type ally.gives.up, version 000"""

from typing import Literal

from pydantic import BaseModel


class AllyGivesUp(BaseModel):
    Reason: str  # This allows us to communicate why we're giving up
    TypeName: Literal["ally.gives.up"] = "ally.gives.up"
    Version: str = "000"



""" List of all the types """

from gwsproto.named_types.actuators_ready import ActuatorsReady
from gwsproto.named_types.admin_analog_dispatch import AdminAnalogDispatch
from gwsproto.named_types.admin_dispatch import AdminDispatch
from gwsproto.named_types.admin_keep_alive import AdminKeepAlive
from gwsproto.named_types.admin_release_control import AdminReleaseControl
from gwsproto.named_types.ally_gives_up import AllyGivesUp
from gwsproto.named_types.atn_bid import AtnBid
from gwsproto.named_types.channel_flatlined import ChannelFlatlined
from gwsproto.named_types.dispatch_contract_go_dormant import DispatchContractGoDormant
from gwsproto.named_types.dispatch_contract_go_live import DispatchContractGoLive
from gwsproto.named_types.energy_instruction import EnergyInstruction
from gwsproto.named_types.events import RemainingElecEvent
from gwsproto.named_types.flo_params import FloParams
from gwsproto.named_types.flo_params_house0 import FloParamsHouse0
from gwsproto.named_types.fsm_event import FsmEvent
from gwsproto.named_types.glitch import Glitch
from gwsproto.named_types.go_dormant import GoDormant
from gwsproto.named_types.ha1_params import Ha1Params
from gwsproto.named_types.hack_oil_off import HackOilOff
from gwsproto.named_types.hack_oil_on import HackOilOn
from gwsproto.named_types.heating_forecast import HeatingForecast
from gwsproto.named_types.latest_price import LatestPrice
from gwsproto.named_types.layout_lite import LayoutLite
from gwsproto.named_types.micro_volts import MicroVolts
from gwsproto.named_types.market_maker_ack import MarketMakerAck
from gwsproto.named_types.new_command_tree import NewCommandTree
from gwsproto.named_types.no_new_contract_warning import NoNewContractWarning
from gwsproto.named_types.pico_missing import PicoMissing
from gwsproto.named_types.price_quantity_unitless import PriceQuantityUnitless
from gwsproto.named_types.remaining_elec import RemainingElec
from gwsproto.named_types.reset_hp_keep_value import ResetHpKeepValue
from gwsproto.named_types.slow_dispatch_contract import SlowDispatchContract
from gwsproto.named_types.scada_params import ScadaParams
from gwsproto.named_types.send_layout import SendLayout
from gwsproto.named_types.set_lwt_control_params import SetLwtControlParams
from gwsproto.named_types.set_target_lwt import SetTargetLwt
from gwsproto.named_types.sieg_loop_endpoint_valve_adjustment import SiegLoopEndpointValveAdjustment
from gwsproto.named_types.sieg_target_too_low import SiegTargetTooLow
from gwsproto.named_types.single_machine_state import SingleMachineState
from gwsproto.named_types.slow_contract_heartbeat import SlowContractHeartbeat
from gwsproto.named_types.snapshot_spaceheat import SnapshotSpaceheat
from gwsproto.named_types.suit_up import SuitUp
from gwsproto.named_types.wake_up import WakeUp
from gwsproto.named_types.weather_forecast import WeatherForecast
from gwsproto.named_types.start_listening_to_atn import StartListeningToAtn
from gwsproto.named_types.stop_listening_to_atn import StopListeningToAtn

__all__ = [
    "ActuatorsReady",
    "AdminAnalogDispatch",
    "AdminDispatch",
    "AdminKeepAlive",
    "AdminReleaseControl",
    "AllyGivesUp",
    "AtnBid",
    "ChannelFlatlined",
    "DispatchContractGoDormant",
    "DispatchContractGoLive",
    "EnergyInstruction",
    "FloParams",
    "FloParamsHouse0",
    "FsmEvent",
    "Glitch",
    "GoDormant",
    "Ha1Params",
    "HackOilOff",
    "HackOilOn",
    "HeatingForecast",
    "LatestPrice",
    "LayoutLite",
    "MarketMakerAck",
    "MicroVolts",
    "NewCommandTree",
    "NoNewContractWarning",
    "PicoMissing",
    "PriceQuantityUnitless",
    "RemainingElec",
    "RemainingElecEvent",
    "ResetHpKeepValue",
    "SlowContractHeartbeat",
    "SlowDispatchContract",
    "ScadaParams",
    "SendLayout",
    "SetLwtControlParams",
    "SetTargetLwt",
    "SiegLoopEndpointValveAdjustment",
    "SiegTargetTooLow",
    "SingleMachineState",
    "SnapshotSpaceheat",
    "StartListeningToAtn",
    "StopListeningToAtn",
    "SuitUp",
    "WakeUp",
    "WeatherForecast",
]



"""Type admin.keep.alive, version 000"""

from typing import Literal, Optional

from pydantic import BaseModel


class AdminKeepAlive(BaseModel):
    AdminTimeoutSeconds: Optional[int] = None
    TypeName: Literal["admin.keep.alive"] = "admin.keep.alive"
    Version: str = "000"



"""Type start.listening.to.atn, version 000"""

from typing import Literal
from gwproto.property_format import LeftRightDotStr
from pydantic import BaseModel


class StartListeningToAtn(BaseModel):
    FromGNodeAlias: LeftRightDotStr
    TypeName: Literal["start.listening.to.atn"] = "start.listening.to.atn"
    Version: str = "000"



"""Type layout.lite, version 002"""

from typing import List, Literal

from gwproto.enums import ActorClass
from gwproto.named_types.data_channel_gt import DataChannelGt
from gwproto.named_types.i2c_multichannel_dt_relay_component_gt import (
    I2cMultichannelDtRelayComponentGt,
)
from gwproto.named_types.pico_flow_module_component_gt import PicoFlowModuleComponentGt
from gwproto.named_types.pico_tank_module_component_gt import PicoTankModuleComponentGt
from gwproto.named_types.spaceheat_node_gt import SpaceheatNodeGt
from gwproto.named_types.synth_channel_gt import SynthChannelGt
from gwproto.property_format import LeftRightDotStr, UTCMilliseconds, UUID4Str
from gwsproto.named_types.ha1_params import Ha1Params
from pydantic import BaseModel, PositiveInt, model_validator
from typing_extensions import Self


class LayoutLite(BaseModel):
    FromGNodeAlias: LeftRightDotStr
    FromGNodeInstanceId: UUID4Str
    MessageCreatedMs: UTCMilliseconds
    MessageId: UUID4Str
    Strategy: str
    ZoneList: List[str]
    TotalStoreTanks: PositiveInt
    ShNodes: List[SpaceheatNodeGt]
    DataChannels: List[DataChannelGt]
    SynthChannels: List[SynthChannelGt]
    TankModuleComponents: List[PicoTankModuleComponentGt]
    FlowModuleComponents: List[PicoFlowModuleComponentGt]
    Ha1Params: Ha1Params
    I2cRelayComponent: I2cMultichannelDtRelayComponentGt
    TypeName: Literal["layout.lite"] = "layout.lite"
    Version: str = "004"

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: Dc Node Consistency. Every AboutNodeName and CapturedByNodeName in a
        DataChannel belongs to an ShNode, and in addition every CapturedByNodeName does
        not have ActorClass NoActor.
        """
        for dc in self.DataChannels:
            if dc.AboutNodeName not in [n.Name for n in self.ShNodes]:
                raise ValueError(
                    f"dc {dc.Name} AboutNodeName {dc.AboutNodeName} not in ShNodes!"
                )
            captured_by_node = next(
                (n for n in self.ShNodes if n.Name == dc.CapturedByNodeName), None
            )
            if not captured_by_node:
                raise ValueError(
                    f"dc {dc.Name} CapturedByNodeName {dc.CapturedByNodeName} not in ShNodes!"
                )
            if captured_by_node.ActorClass == ActorClass.NoActor:
                raise ValueError(
                    f"dc {dc.Name}'s CatpuredByNode cannot have ActorClass NoActor!"
                )
        return self

    @model_validator(mode="after")
    def check_axiom_2(self) -> Self:
        """
        Node Handle Hierarchy Consistency. Every ShNode with a handle containing at least
        two words (separated by '.') has an immediate boss: another ShNode whose handle
        matches the original handle minus its last word.
        """
        existing_handles = {get_handle(node) for node in self.ShNodes}
        for node in self.ShNodes:
            handle = get_handle(node)
            if "." in handle:
                boss_handle = ".".join(handle.split(".")[:-1])
                if boss_handle not in existing_handles:
                    raise ValueError(
                        f"node {node.Name} with handle {handle} missing"
                        " its immediate boss!"
                    )
        return self


def get_handle(node: SpaceheatNodeGt) -> str:
    if node.Handle:
        return node.Handle
    return node.Name



"""Type energy.instruction, version 000"""

from typing import Literal

from gwproto.property_format import LeftRightDotStr, UTCMilliseconds, UTCSeconds
from pydantic import BaseModel, PositiveInt, StrictInt, field_validator, model_validator
from typing_extensions import Self


class EnergyInstruction(BaseModel):
    FromGNodeAlias: LeftRightDotStr
    SlotStartS: UTCSeconds
    SlotDurationMinutes: PositiveInt
    SendTimeMs: UTCMilliseconds
    AvgPowerWatts: StrictInt
    TypeName: Literal["energy.instruction"] = "energy.instruction"
    Version: str = "000"

    @field_validator("SlotStartS")
    @classmethod
    def check_slot_start_s(cls, v: int) -> int:
        """
        Axiom 1: SlotStartS should fall on the top of 5 minutes
        """
        if v % 300 != 0:
            raise ValueError(
                f"Axiom 1: SlotStartS should fall on the top of 5 minutes. But got {v % 300} seconds out"
            )
        return v

    @model_validator(mode="after")
    def check_axiom_2(self) -> Self:
        """
        Axiom 2: SendTime within 10 seconds of SlotStart.
        SendTimeMs should be no more than 10 seconds after SlotStartS
        """
        send_time_s = self.SendTimeMs / 1000
        if send_time_s > self.SlotStartS + 10:
            raise ValueError(
                f"Axiom 2: SendTime within 10 seconds of SlotStart. Got {round(send_time_s - self.SlotStartS, 2)}"
            )
        return self



from enum import auto
from typing import List

from gw.enums import GwStrEnum


class AtomicAllyState(GwStrEnum):
    Dormant = auto()
    Initializing = auto()
    HpOnStoreOff = auto()
    HpOnStoreCharge = auto()
    HpOffStoreOff = auto()
    HpOffStoreDischarge = auto()
    HpOffOilBoilerTankAquastat = auto()

    @classmethod
    def default(cls) -> "AtomicAllyState":
        return cls.Dormant

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]
    
    @classmethod
    def enum_name(cls) -> str:
        return "atomic.ally.state"

    @classmethod
    def enum_version(cls) -> str:
        return "000"



import json
from pathlib import Path
from typing import Any, List, Literal, Optional

from gw.errors import DcError
from gwproto.enums import ActorClass
from gwproto.data_classes.components import Component
from gwproto.data_classes.data_channel import DataChannel
from gwproto.data_classes.hardware_layout import (
    HardwareLayout,
    LoadArgs,
    LoadError,
)

from gwsproto.data_classes.house_0_names import H0CN, H0N
from gwsproto.enums import FlowManifoldVariant, HomeAloneStrategy

from gwproto.data_classes.sh_node import ShNode
from gwproto.data_classes.synth_channel import SynthChannel
from gwproto.default_decoders import (
    CacDecoder,
    ComponentDecoder,
)
from gwproto.named_types import ComponentAttributeClassGt

class House0LoadArgs(LoadArgs):
    flow_manifold_variant: FlowManifoldVariant
    use_sieg_loop: bool
class House0Layout(HardwareLayout):
    zone_list: List[str]
    total_store_tanks: int


    def __init__(  # noqa: PLR0913
        self,
        layout: dict[Any, Any],
        *,
        cacs: dict[str, ComponentAttributeClassGt],  # by id
        components: dict[str, Component],  # by id
        nodes: dict[str, ShNode],  # by name
        data_channels: dict[str, DataChannel],  # by name
        synth_channels: dict[str, SynthChannel],
        flow_manifold_variant: FlowManifoldVariant = FlowManifoldVariant.House0,
        use_sieg_loop: bool = False,
    ) -> None:
        super().__init__(
            layout=layout,
            cacs=cacs,
            components=components,
            nodes=nodes,
            data_channels=data_channels,
            synth_channels=synth_channels,
        )
        self.flow_manifold_variant = flow_manifold_variant
        self.use_sieg_loop = use_sieg_loop

        # Bolted on right now
        required_keys = ["ZoneList", "TotalStoreTanks"]
        for key in required_keys:
            if key not in layout:
                raise DcError(f"House0 requires {key}!")


        self.zone_list = layout["ZoneList"]
        self.total_store_tanks = layout["TotalStoreTanks"]

        self.channel_names = H0CN(self.total_store_tanks, self.zone_list)
        if not isinstance(self.total_store_tanks, int):
            raise TypeError("TotalStoreTanks must be an integer")
        if not 1 <= self.total_store_tanks <= 6:
            raise ValueError("Must have between 1 and 6 store tanks")
        if not isinstance(self.zone_list, List):
            raise TypeError("ZoneList must be a list")
        if not 1 <= len(self.zone_list) <= 6:
            raise ValueError("Must have between 1 and 6 store zones")
        self.h0n = H0N(self.total_store_tanks, self.zone_list)

    @classmethod
    def validate_house0(  # noqa: C901
        cls,
        load_args: House0LoadArgs,
        *,
        raise_errors: bool,
        errors: Optional[list[LoadError]] = None,
    ) -> None:
        nodes = load_args["nodes"]
        data_channels = load_args["data_channels"]
        errors_caught = []
        flow_manifold_variant = load_args["flow_manifold_variant"]
        use_sieg_loop = load_args["use_sieg_loop"]

        # Can't use the siegenthaler loop in the code if it isn't in the plumbing
        if use_sieg_loop and flow_manifold_variant != FlowManifoldVariant.House0Sieg:
            raise DcError("Cannot use Sieg Loop when FlowManifoldVariant is not House0Sieg!")

        # Make sure sieg relays, sieg flow and sieg temp nodes and channels exist
        if flow_manifold_variant == FlowManifoldVariant.House0Sieg:
            try:
                cls.check_house0_sieg_manifold(data_channels)
            except Exception as e:
                if raise_errors:
                    raise
                errors_caught.append(LoadError("hardware.layout", nodes, e))


        if use_sieg_loop: # HpBoss and SiegLoop need to be actors
            try:
                cls.check_actors_when_using_sieg_loop(nodes)
            except Exception as e:
                if raise_errors:
                    raise
                errors_caught.append(LoadError("hardware.layout", nodes, e))
        else: # HpBoss and SiegLoop should NOT be actors
            try:
                cls.check_actors_when_not_using_sieg_loop(nodes)
            except Exception as e:
                if raise_errors:
                    raise
                errors_caught.append(LoadError("hardware.layout", nodes, e))
    @classmethod
    def check_house0_sieg_manifold(cls, channels: dict[str, DataChannel]) -> None:
        # if H0CN.sieg_cold not in channels.keys():
        #     raise DcError(f"Need {H0CN.sieg_cold} channel with House0Sieg flow manifold variant")
        # if H0CN.sieg_flow not in channels.keys():
        #     raise DcError(f"Need {H0CN.sieg_flow} channel with House0Sieg flow manifold variant")
        if H0CN.hp_loop_on_off_relay_state not in channels.keys():
            raise DcError(f"Need {H0CN.hp_loop_on_off_relay_state} channel with House0Sieg flow manifold variant")
        if H0CN.hp_loop_keep_send_relay_state not in channels.keys():
            raise DcError(f"Need {H0CN.hp_loop_keep_send_relay_state} channel with House0Sieg flow manifold variant")

    @classmethod
    def check_actors_when_using_sieg_loop(cls, nodes: dict[str, ShNode]) -> None:
        if H0N.sieg_loop not in nodes.keys():
            raise DcError(f"Need a SiegLoop actor when using sieg loop!")
        sieg_loop = nodes[H0N.sieg_loop]
        if sieg_loop.actor_class != ActorClass.SiegLoop:
            raise DcError(f"SiegLoop actor {sieg_loop.name} shoud have actor class SiegLoop, not {sieg_loop.actor_class}")
        if H0N.hp_boss not in nodes.keys():
            raise DcError(f"Need HpBoss actor when using sieg loop!")
        hp_boss = nodes[H0N.hp_boss]
        if hp_boss.actor_class != ActorClass.HpBoss:
            raise DcError(f"HpBoss actor {hp_boss.name} shoud have actor class HpBoss, not {hp_boss.actor_class}")

    @classmethod
    def check_actors_when_not_using_sieg_loop(cls, nodes: dict[str, ShNode]) -> None:
        if H0N.sieg_loop in nodes.keys():
            raise DcError(f"If not using sieg loop, should not have node {H0N.sieg_loop}!")

    @property
    def ha_strategy(self) -> str:
        """Returns the current home alone strategy"""
        # Could be stored as a property or derived from a node
        ha_node = self.nodes.get(H0N.home_alone)
        return HomeAloneStrategy(HomeAloneStrategy(getattr(ha_node, "Strategy", None)))
    
    @property
    def actuators(self) -> List[ShNode]:
        return self.relays + self.zero_tens
    
    @property
    def relays(self) -> List[ShNode]:
        return [
            node for node in self.nodes.values()
            if node.ActorClass == ActorClass.Relay
        ]
    
    @property
    def zero_tens(self) -> List[ShNode]:
        return [
            node for node in self.nodes.values()
            if node.ActorClass == ActorClass.ZeroTenOutputer
        ]


    # overwrites base class to return correct object
    @classmethod
    def load(  # noqa: PLR0913
        cls,
        layout_path: Path | str,
        *,
        included_node_names: Optional[set[str]] = None,
        raise_errors: bool = True,
        errors: Optional[list[LoadError]] = None,
        cac_decoder: Optional[CacDecoder] = None,
        component_decoder: Optional[ComponentDecoder] = None,
    ) -> "House0Layout":
        with Path(layout_path).open() as f:
            layout = json.loads(f.read())
        return cls.load_dict(
            layout,
            included_node_names=included_node_names,
            raise_errors=raise_errors,
            errors=errors,
            cac_decoder=cac_decoder,
            component_decoder=component_decoder,
        )

    # overwrites base class to return correct object
    @classmethod
    def load_dict(  # noqa: PLR0913
        cls,
        layout: dict[Any, Any],
        *,
        included_node_names: Optional[set[str]] = None,
        raise_errors: bool = True,
        errors: Optional[list[LoadError]] = None,
        cac_decoder: Optional[CacDecoder] = None,
        component_decoder: Optional[ComponentDecoder] = None,
    ) -> "House0Layout":
        if errors is None:
            errors = []
        cacs = cls.load_cacs(
            layout=layout,
            raise_errors=raise_errors,
            errors=errors,
            cac_decoder=cac_decoder,
        )
        components = cls.load_components(
            layout=layout,
            cacs=cacs,
            raise_errors=raise_errors,
            errors=errors,
            component_decoder=component_decoder,
        )
        nodes = cls.load_nodes(
            layout=layout,
            components=components,
            raise_errors=raise_errors,
            errors=errors,
            included_node_names=included_node_names,
        )
        data_channels = cls.load_data_channels(
            layout=layout,
            nodes=nodes,
            raise_errors=raise_errors,
            errors=errors,
        )
        synth_channels = cls.load_synth_channels(
            layout=layout,
            nodes=nodes,
            raise_errors=raise_errors,
            errors=errors,
        )
        load_args: House0LoadArgs = {
            "cacs": cacs,
            "components": components,
            "nodes": nodes,
            "data_channels": data_channels,
            "synth_channels": synth_channels,
            "flow_manifold_variant": FlowManifoldVariant(layout.get("FlowManifoldVariant", "House0")),
            "use_sieg_loop": bool(layout.get("UseSiegLoop", False))
        }
        cls.resolve_links(
            load_args["nodes"],
            load_args["components"],
            raise_errors=raise_errors,
            errors=errors,
        )
        cls.validate_layout(load_args, raise_errors=raise_errors, errors=errors)
        cls.validate_house0(load_args, raise_errors=raise_errors, errors=errors)
        return House0Layout(layout, **load_args)

    @property
    def home_alone(self) -> ShNode:
        return self.node(H0N.home_alone)
    
    @property
    def auto_node(self) -> ShNode:
        return self.node(H0N.auto)
    
    @property
    def atomic_ally(self) -> ShNode:
        return self.node(H0N.atomic_ally)
    
    @property
    def atn(self) -> ShNode:
        return self.node(H0N.atn)
    
    @property
    def pico_cycler(self) -> ShNode:
        return self.node(H0N.pico_cycler)

    @property
    def vdc_relay(self) -> ShNode:
        return self.node(H0N.vdc_relay)

    @property
    def tstat_common_relay(self) -> ShNode:
        return self.node(H0N.tstat_common_relay)

    @property
    def charge_discharge_relay(self) -> ShNode:
        return self.node(H0N.store_charge_discharge_relay)#

    def scada2_gnode_name(self) -> str:
        return f"{self.scada_g_node_alias}.{H0N.secondary_scada}"

def deserialize_house0_load_args(data: dict) -> House0LoadArgs:
    valid_keys = set(House0LoadArgs.__annotations__.keys())

    # Validate the FlowManifoldVariant
    data["FlowManifoldVariant"] = FlowManifoldVariant(data.get("FlowManifoldVariant", "House0"))
    # Validate use_sieg_loop
    data["UseSiegLoop"] = bool(data.get("UseSiegLoop", False))
    # TypedDict expects a regular dictionary, so we just pass it in
    return House0LoadArgs(**data)




from enum import auto
from typing import List

from gw.enums import GwStrEnum


class PicoCyclerState(GwStrEnum):
    """
    
    Values:
      - Dormant
      - PicosLive
      - RelayOpening
      - RelayOpen
      - RelayClosing
      - PicosRebooting
      - AllZombies

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#picocyclerstate)
    """

    Dormant = auto()
    PicosLive = auto()
    RelayOpening = auto()
    RelayOpen = auto()
    RelayClosing = auto()
    PicosRebooting = auto()
    AllZombies = auto()

    @classmethod
    def default(cls) -> "PicoCyclerState":
        return cls.PicosLive

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "pico.cycler.state"

    @classmethod
    def enum_version(cls) -> str:
        return "000"



"""Type sieg.loop.endpoint.valve.adjustment, version 000"""
import time
from typing import Literal

from gwproto.property_format import HandleName, UTCMilliseconds
from pydantic import BaseModel, Field, PositiveInt, StrictInt, model_validator
from typing_extensions import Self

class SiegLoopEndpointValveAdjustment(BaseModel):
    """
    Used to gauarantee that the Siegenthaler valve is fully closed or fully open.
    If the HpKeepPercent is 100% (and matches actual), the SiegLoop actor pushes
    the Siegenthaler valve to "keep more". If 0%, it 
    pushes the valve to "keep less". 
    """ 
    FromHandle: HandleName
    ToHandle: HandleName
    HpKeepPercent: StrictInt
    Seconds: PositiveInt
    CreatedMs: UTCMilliseconds = Field(default_factory=lambda: int(time.time() * 1000))
    TypeName: Literal["sieg.loop.endpoint.valve.adjustment"] = "sieg.loop.endpoint.valve.adjustment"
    Version: str = "000"

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: HpKeepPercent is 0 or 100
        """
        if self.HpKeepPercent not in [0, 100]:
            raise ValueError(f"HpKeepPercent must be 0 or 100, not {self.HpKeepPercent}")
        return self




"""Type weather.forecast, version 000"""
import time
import uuid
from typing import List, Literal

from gwproto.property_format import  LeftRightDotStr, UTCSeconds, UUID4Str
from pydantic import BaseModel, Field, model_validator
from typing_extensions import Self


class WeatherForecast(BaseModel):
    """
    Weather Forecast Raw.

    A type meant to be used for weather forecast data from a direct third-party source.
    """

    FromGNodeAlias: LeftRightDotStr
    WeatherChannelName: LeftRightDotStr
    Time: List[UTCSeconds]
    OatF: List[float]
    WindSpeedMph: List[float]
    WeatherUid: UUID4Str = Field(default_factory=lambda: str(uuid.uuid4()))
    ForecastCreatedS: UTCSeconds = Field(default_factory=lambda: int(time.time()))
    TypeName: Literal["weather.forecast"] = "weather.forecast"
    Version: str = "000"

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: Length of all the lists (Time, OatF, WindspeedMph) must be the same.

        """
        if len(self.OatF) != len(self.Time):
            raise ValueError(f"self.OatFshould be a list of length {len(self.Time)}!")
        if len(self.WindSpeedMph) != len(self.Time):
            raise ValueError(f"self.OatF should be a list of length {len(self.Time)}!")

        return self

    @model_validator(mode="after")
    def check_axiom_2(self) -> Self:
        """
        Axiom 2: ForecastCreatedS is less than the first second in Time

        """
        if len(self.Time) > 0:
            if self.ForecastCreatedS > self.Time[0]:
                raise ValueError(
                    "Forecast should not be created after beginning of forecasted time!"
                )

        return self



from typing import Literal
import time

from pydantic import BaseModel, field_validator, model_validator, PositiveInt, StrictInt
from gwproto.property_format import UUID4Str, UTCSeconds,  LeftRightDotStr
from typing_extensions import Self

class SlowDispatchContract(BaseModel):
    """Represents a dispatch contract between Atn and Scada"""
    ScadaAlias: LeftRightDotStr
    StartS: UTCSeconds
    DurationMinutes: PositiveInt
    AvgPowerWatts:StrictInt
    OilBoilerOn: bool
    ContractId: UUID4Str
    TypeName: Literal["slow.dispatch.contract"] = "slow.dispatch.contract"
    Version: str = "000"

    @field_validator("StartS")
    @classmethod
    def check_start_s(cls, v: int) -> int:
        """
        Axiom 1: StartS should fall on the top of 5 minutes
        """
        if v % 300 != 0:
            raise ValueError(
                f"Axiom 1: SlotStartS should fall on the top of 5 minutes. But got {v % 300} seconds out"
            )
        return v

    @model_validator(mode="after")
    def _check_axiom_1(self) -> Self:
        """OilBoilerOn -> AvgPowerWatts is 0
        """
        if self.OilBoilerOn:
            if self.AvgPowerWatts > 0:
                raise ValueError("OilBoilerOn -> AvgPowerWatts is 0")
        return self

    def contract_end_s(self) -> int:
        return self.StartS + self.DurationMinutes * 60




from enum import auto
from typing import List

from gw.enums import GwStrEnum


class MainAutoState(GwStrEnum):
    """
    
    Values:
      - HomeAlone
      - Atn
      - Dormant

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#mainautostate)
    """

    HomeAlone = auto()
    Atn = auto()
    Dormant = auto()

    @classmethod
    def default(cls) -> "MainAutoState":
        return cls.HomeAlone

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "main.auto.state"

    @classmethod
    def enum_version(cls) -> str:
        return "000"



"""Type actuators.ready version 000"""
from typing import Literal

from pydantic import BaseModel


class ActuatorsReady(BaseModel):
    """
    Used to sequence inbound messages to actuators
    """
    TypeName: Literal["actuators.ready"] = "actuators.ready"
    Version: str = "000"



"""Type suit.up, version 000"""

from typing import Literal

from gwproto.property_format import SpaceheatName
from pydantic import BaseModel


class SuitUp(BaseModel):
    """ """

    ToNode: SpaceheatName
    FromNode: SpaceheatName
    TypeName: Literal["suit.up"] = "suit.up"
    Version: str = "000"



"""Type send.layout, version 000"""

from typing import Literal

from gwproto.property_format import LeftRightDotStr, SpaceheatName
from pydantic import BaseModel


class SendLayout(BaseModel):
    FromGNodeAlias: LeftRightDotStr
    FromName: SpaceheatName
    ToName: SpaceheatName
    TypeName: Literal["send.layout"] = "send.layout"
    Version: str = "000"



"""Type flo.params.house0, version 000"""
import time
import uuid
from typing import List, Literal, Optional

from gwsproto.enums import MarketPriceUnit
from gwproto.property_format import LeftRightDotStr, UTCSeconds, UUID4Str
from pydantic import BaseModel, ConfigDict, Field, PositiveInt, StrictInt


class FloParamsHouse0(BaseModel):
    GNodeAlias: LeftRightDotStr
    FloParamsUid: UUID4Str = Field(default_factory=lambda: str(uuid.uuid4()))
    TimezoneStr: str = "America/New_York"
    StartUnixS: UTCSeconds
    HorizonHours: PositiveInt = 48
    NumLayers: PositiveInt = 24
    # Equipment
    StorageVolumeGallons: PositiveInt = 360
    StorageLossesPercent: float = 0.5
    HpMinElecKw: float = -0.5
    HpMaxElecKw: float = 11
    CopIntercept: float = 1.02
    CopOatCoeff: float = 0.0257
    CopLwtCoeff: float = 0
    CopMin: float = 1.4
    CopMinOatF: float = 15
    HpTurnOnMinutes: int = 10
    # Initial state
    InitialTopTempF: StrictInt
    InitialMiddleTempF: StrictInt
    InitialBottomTempF: StrictInt
    InitialThermocline1: StrictInt
    InitialThermocline2: StrictInt
    HpIsOff: bool = False
    BufferAvailableKwh: float = 0
    HouseAvailableKwh: float = 0
    # Forecasts
    LmpForecast: Optional[List[float]] = None
    DistPriceForecast: Optional[List[float]] = None
    RegPriceForecast: Optional[List[float]] = None
    PriceForecastUid: UUID4Str = Field(default_factory=lambda: str(uuid.uuid4()))
    OatForecastF: Optional[List[float]] = None
    WindSpeedForecastMph: Optional[List[float]] = None
    WeatherUid: UUID4Str = Field(default_factory=lambda: str(uuid.uuid4()))
    # House parameters
    AlphaTimes10: StrictInt
    BetaTimes100: StrictInt
    GammaEx6: StrictInt
    IntermediatePowerKw: float
    IntermediateRswtF: StrictInt
    DdPowerKw: float
    DdRswtF: StrictInt
    DdDeltaTF: StrictInt
    MaxEwtF: StrictInt
    PriceUnit: MarketPriceUnit = MarketPriceUnit.USDPerMWh
    ParamsGeneratedS: UTCSeconds = Field(default_factory=lambda: int(time.time()))
    # Flo type
    FloAlias: str = "Winter.Oak"
    FloGitCommit: str = "76e267b"
    TypeName: Literal["flo.params.house0"] = "flo.params.house0"
    Version: str = "003"

    model_config = ConfigDict(extra="allow", use_enum_values=True)



from enum import auto
from gw.enums import GwStrEnum
from typing import List

class ContractStatus(GwStrEnum):
    """Status enum for tracking contract lifecycle

    Values:
      - Created # Initial contract proposal
      - Received # Counter-party acknowledges receipt
      - Confirmed # Counter-party accepts terms
      - Active # Contract currently in force
      - TerminatedByAtn # Contract ended early by Atn
      - TerminatedByScada # Contract ended early by Scada
      - CompletedSuccess # Contract ran full duration and terms met
      - CompletedFailureByScada # Contract ran full duration, terms not met by Scada
      - CompletedFailureByAtn # Contract ran full duration, terms not met by Atn

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
    """
    Created = auto()
    Received = auto() 
    Confirmed = auto()
    Active = auto()
    TerminatedByAtn = auto()
    TerminatedByScada = auto()
    CompletedUnknownOutcome = auto()
    CompletedSuccess = auto()
    CompletedFailureByScada = auto()
    CompletedFailureByAtn = auto()

    @classmethod
    def default(cls) -> "ContractStatus":
        return cls.Created

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "contract.status"

    @classmethod 
    def enum_version(cls) -> str:
        return "000"


"""Type set.target.lwt, version 000"""
import time
from typing import Literal

from gwproto.property_format import HandleName, UTCMilliseconds
from pydantic import BaseModel, PositiveInt, Field, model_validator
from typing_extensions import Self

class SetTargetLwt(BaseModel):
    FromHandle: HandleName
    ToHandle: HandleName
    TargetLwtF: PositiveInt
    CreatedMs: UTCMilliseconds = Field(default_factory=lambda: int(time.time() * 1000))
    TypeName: Literal["set.target.lwt"] = "set.target.lwt"
    Version: str = "000"

    @model_validator(mode="after")
    def check_axiom_1(self) -> Self:
        """
        Axiom 1: FromHandle is boss of ToHandle
        """
        immediate_boss = ".".join(self.ToHandle.split(".")[:-1])
        if immediate_boss != self.FromHandle:
            raise ValueError(
                f"FromHandle {self.FromHandle} must be immediate boss of ToHandle {self.ToHandle}"
            )
        return self


from enum import auto
from typing import List

from gw.enums import GwStrEnum


class MainAutoEvent(GwStrEnum):
    """
    
    Values:
      - DispatchContractLive # HomeAlone -> Atn
      - ContractGracePeriodEnds # Atn -> HomeAlone
      - AtnReleasesControl # Atn -> HomeAlone
      - AllyGivesUp # Atn -> HomeAlone
      - AutoGoesDormant # Atn or HomeAlone -> Dormant
      - AutoWakesUp # Dormant -> HomeAlone

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#mainautoevent)
    """

    DispatchContractLive = auto()
    ContractGracePeriodEnds = auto()
    AtnReleasesControl = auto()
    AllyGivesUp = auto()
    AutoGoesDormant = auto()
    AutoWakesUp = auto()

    @classmethod
    def default(cls) -> "MainAutoEvent":
        return cls.AutoWakesUp

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "main.auto.event"

    @classmethod
    def enum_version(cls) -> str:
        return "001"



"""Type hack.oil.off, version 000"""

from typing import Literal

from pydantic import BaseModel


class HackOilOff(BaseModel):
    TypeName: Literal["hack.oil.off"] = "hack.oil.off"
    Version: str = "000"



"""Type new.command.tree, version 000"""

from typing import List, Literal

from gwproto.named_types.spaceheat_node_gt import SpaceheatNodeGt
from gwproto.property_format import LeftRightDotStr, UTCMilliseconds
from pydantic import BaseModel, field_validator


class NewCommandTree(BaseModel):
    FromGNodeAlias: LeftRightDotStr
    ShNodes: List[SpaceheatNodeGt]
    UnixMs: UTCMilliseconds
    TypeName: Literal["new.command.tree"] = "new.command.tree"
    Version: str = "000"

    @field_validator("ShNodes")
    @classmethod
    def check_sh_nodes(cls, v: List[SpaceheatNodeGt]) -> List[SpaceheatNodeGt]:
        """
            Axiom 1: ShNode hierarchy consistencies.
            Handles and ActorHierarchyNames are closed under prefixes. That is, for every handle in
        the set of handles, all of its prefixes (as generated by handle.split(".")[:n] for any n)
        are also in the set of handles.
        """
        # Implement Axiom(s)
        return v



"""Type admin.dispatch, version 000"""

from typing import Literal

from gwproto.named_types import AnalogDispatch

from pydantic import BaseModel, StrictInt


class AdminAnalogDispatch(BaseModel):
    Dispatch: AnalogDispatch
    TimeoutSeconds: StrictInt
    TypeName: Literal["admin.analog.dispatch"] = "admin.analog.dispatch"
    Version: str = "000"



"""Type admin.release.control, version 000"""

from typing import Literal

from pydantic import BaseModel


class AdminReleaseControl(BaseModel):
    """ """

    TypeName: Literal["admin.release.control"] = "admin.release.control"
    Version: str = "000"



from enum import auto
from typing import List

from gw.enums import GwStrEnum


class PicoCyclerEvent(GwStrEnum):
    """
    
    Values:
      - WakeUp
      - GoDormant
      - PicoMissing
      - ConfirmOpened
      - StartClosing
      - ConfirmClosed
      - ConfirmRebooted
      - ShakeZombies
      - RebootDud

    For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#picocyclerevent)
    """

    WakeUp = auto()
    GoDormant = auto()
    PicoMissing = auto()
    ConfirmOpened = auto()
    StartClosing = auto()
    ConfirmClosed = auto()
    ConfirmRebooted = auto()
    ShakeZombies = auto()
    RebootDud = auto()

    @classmethod
    def default(cls) -> "PicoCyclerEvent":
        return cls.ConfirmRebooted

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]

    @classmethod
    def enum_name(cls) -> str:
        return "pico.cycler.event"

    @classmethod
    def enum_version(cls) -> str:
        return "000"



"""Type flo.params, version 001"""

from typing import Literal

from gwproto.property_format import LeftRightDotStr, UTCSeconds, UUID4Str
from pydantic import BaseModel, ConfigDict, field_validator


class FloParams(BaseModel):
    """
    Base class for Forward Looking Optimizer params.

    Derived classes are expected to have TypeNames enforced as literals that start with flo.params.
    E.g. flo.params.brickstorageheater. This container is used for sending messages that include
    flo.params (i.e, flo.params.report

    [More info](https://gridworks-atn.readthedocs.io/en/latest/flo.html#flo-params)
    """

    GNodeAlias: LeftRightDotStr
    FloParamsUid: UUID4Str
    HomeCity: str
    TimezoneString: str
    StartUnixS: UTCSeconds
    TypeName: Literal["flo.params"] = "flo.params"
    Version: str = "001"

    model_config = ConfigDict(extra="allow")

    @field_validator("StartUnixS")
    @classmethod
    def check_start_unix_s(cls, v: int) -> int:
        """
        Axiom 1: StartS is 0 mod 300
        """
        # Implement Axiom(s)
        return v



from enum import auto
from gw.enums import GwStrEnum

class HpModel(GwStrEnum):
    LgHighTempHydroKitPlusMultiV = auto()  
    SamsungFourTonneHydroKit = auto()   
    SamsungFiveTonneHydroKit = auto()

    @classmethod
    def enum_name(cls) -> str:
        return "hp.model"
